/* Generation of combinations of p elements out of n in lexicographic order.
   Adaptation of the algorithm given page 178 of the fxtbook.
   The combination {e_0,..,e_{p-1}} with e_0 < ... < e_{p-1} is stored in the
   array c of size p such that c[i] = e_i. */

#include "comb.h"

/*@ requires 1 < p <= n;
  @ requires \valid(c+(0..p-1));
  @ requires 1 <= rev <= p-1;
  @ requires is_comb(c,p,n);
  @ requires is_le(c,0,p,p,n);
  @ assigns c[rev..p-1];
  @ ensures is_comb(c,p,n);
  @ ensures is_le(c,0,p,p,n); */
void fillEnd(int c[], int rev, int p, int n) {
 int k;
 /*@ loop invariant l1: rev <= k <= p;
   @ loop invariant l2: is_lt_next(c,rev-1,k-1);
   @ loop invariant l3: is_le(c,rev-1,k,p,n);
   @ loop assigns k, c[rev..p-1];
   @ loop variant p-k; */
 for (k = rev; k < p; k++) {
   c[k] = c[k-1]+1;
 }
 /*@ assert a4: is_lt_next(c,rev-1,p-1); */
 /*@ assert a5: is_le(c,rev-1,p,p,n); */
 /*@ assert b6: is_lt(c,0,rev,n); */
 /*@ assert b7: is_lt(c,rev+1,p,n); */
 /*@ assert b8: is_lt(c,0,p,n); */
 /*@ assert b9: is_lt_next(c,0,p-1); */
}

/*@ requires 1 < p <= n;
  @ requires \valid(c+(0..p-1));
  @ requires is_comb(c,p,n);
  @ requires is_le(c,0,p,p,n);
  @ assigns c[0..p-1];
  @ ensures is_comb(c,p,n);
  @ ensures is_le(c,0,p,p,n); */
int nextComb(int c[], int p, int n) {
 int k,rev;

 if (p == n) {
  return 0;
 }
 /*@ assigns c[p-1];
   @ ensures e1: c[p-1] == \old(c[p-1]) || c[p-1] == \old(c[p-1])+1;
   @ ensures e2: c[p-1] < n;
   @ ensures e3: c[p-1] >= \old(c[p-1]); */
 if (c[p-1] < n-1) { // easy case: highest elt != highest possible value
  c[p-1] = c[p-1]+1;
  return 1;
 }
 rev = p-1;
 /*@ loop invariant 0 <= rev <= p-1;
   @ loop invariant l1: is_le(c,rev,p,p,n);
   @ loop assigns rev;
   @ loop variant rev; */
 while (rev >= 1 && c[rev-1] >= c[rev]-1) {
  rev--;
 }
 if (rev < 1) {
  return 0;
 }
 /*@ assert b1: is_lt_next(c,0,p-1); */
 /*@ assert a1: c[rev-1] < c[rev]-1; */
 /*@ assert a2: c[rev-1] <= n-1-(p-1-rev)-2; */

 /*@ assigns c[rev-1];
   @ ensures e4: c[rev-1] == \old(c[rev-1])+1;
   @ ensures e5: c[rev-1] < n; */
 c[rev-1] = c[rev-1]+1;
 /*@ assert b2: is_lt_next(c,0,rev); */
 /*@ assert b3: is_lt_next(c,rev+1,p-1); */
 /*@ assert b4: is_lt_next(c,0,p-1); */
 /*@ assert a3: is_lt(c,0,rev-1,n); */
 /*@ assert a4: is_lt(c,rev,p,n); */
 /*@ assert a5: is_lt(c,0,p,n); */
 /*@ assert a6: is_le(c,0,rev-1,p,n); */
 /*@ assert a7: is_le(c,rev-1,rev,p,n); */
 /*@ assert a8: is_le(c,rev,p,p,n); */
 /*@ assert a9: is_le(c,0,p,p,n); */
 /*@ assert a10: is_comb(c,p,n); */
 fillEnd(c,rev,p,n);
 return 1;
}

/*@ requires 1 <= p <= n;
  @ requires \valid(c+(0..p-1));
  @ assigns c[0..p-1];
  @ ensures is_id(c,p);
  @ ensures is_comb(c,p,n); */
void firstComb(int c[], int p, int n) {
 int k;
 /*@ loop invariant 0 <= k <= p;
   @ loop invariant is_id(c,k);
   @ loop assigns k, c[0..p-1];
   @ loop variant p-k; */
 for (k = 0; k < p; k++) {
  c[k] = k;
 }
}
