/* Generation of combinations of p elements out of n in the lexicographic order
   induced by the precedence ? < ?.
   Here, by filtering among all functions from [0..p-1] to [0..n-1].
   All is proved with frama-c Neon and
   - alt-Ergo 0.95.1 in .944 s
   - cvc3 2.4.1 in .23 s */

#include "../fct/fct.h"

/*@ predicate is_inc(int *a, integer n) = \forall integer i; 1 <= i < n ==> a[i] > a[i-1];
  @ predicate is_comb(int *a, integer p, integer n) = is_fct(a,p,n) && is_inc(a,p); */

/* Returns 1 when the array a is a combination
   of p elements out of n, and 0 otherwise. */
/*@ requires p > 0 && n >= p && \valid(a+(0..p-1));
  @ requires is_fct(a,p,n);
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_comb(a,p,n); */
int isComb(int a[], int p, int n) {
 int i;
 
 /*@ loop invariant 1 <= i <= p;
   @ loop invariant is_inc(a,i);
   @ loop assigns i;
   @ loop variant p-i; */
 for (i = 1; i < p; i++) {
  if (a[i] <= a[i-1]) {
   return 0;
  } 
 }
 return 1;
}

/*@ requires p > 0 && n >= p && \valid(a+(0..p-1));
  @ assigns a[0..p-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_comb(a,p,n); */
int firstComb(int a[], int p, int n) {
 int tmp;
 
 tmp = firstFct(a,p,n);
 /*@ loop invariant tmp != 0 ==> is_fct(a,p,n);
   @ loop assigns a[0..p-1],tmp; */
 while (tmp != 0 && isComb(a,p,n) == 0) {
  tmp = nextFct(a,p,n);
 } 
 if (tmp == 0) { return 0; }
 return 1;
}


/*@ requires p > 0 && n >= p && \valid(a+(0..p-1));
  @ requires is_comb(a,p,n);
  @ assigns a[0..p-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_comb(a,p,n); */
int nextComb(int a[], int p, int n) {
 int tmp;
 
 /*@ loop invariant is_fct(a,p,n);
   @ loop assigns a[0..p-1],tmp; */
 do {
  tmp = nextFct(a,p,n);
 } while (tmp != 0 && isComb(a,p,n) == 0);
 if (tmp == 0) { return 0; }
 return 1;
}
