/* Generation algorithm for RGFs (see also fxtbook/rgf/).
 * Here, by filtering among all endofunctions of [0..n-1].
 * Proved with Frama-C Neon and Alt-Ergo 0.95.1. */

#include "../fct/fct.h"
#include "../endofct/endofct.h"

/*@ predicate is_le_pred(int *a, integer n) =
  @  (\forall integer i; 1 <= i < n ==> a[i] <= a[i-1]+1); */

/*@ predicate is_RGF(int *a, integer n) =
  @  is_endofct(a,n) && a[0] == 0
  @  && is_le_pred(a,n); */

/* Returns 1 when the array a is an RGF, and
 * 0 otherwise. */
/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires is_endofct(a,n);
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_RGF(a,n); */
int isRGF(int a[], int n) {
 int i;
 
 if (a[0] != 0) return 0;
 /*@ loop invariant 1 <= i <= n;
   @ loop invariant is_le_pred(a,i);
   @ loop assigns i;
   @ loop variant n-i; */
 for (i = 1; i < n; i++)
  if (a[i] > a[i-1]+1) return 0;
 return 1;
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires is_RGF(a,n);
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_RGF(a,n); */
int nextRGF(int a[], int n) {
 int tmp;

 /*@ loop invariant is_endofct(a,n);
   @ loop assigns a[0..n-1],tmp; */
 do {
  tmp = nextEndofct(a,n);
 } while (tmp != 0 && isRGF(a,n) == 0);
 if (tmp == 0) { return 0; }
 return 1;
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_RGF(a,n); */
int firstRGF(int a[], int n) {
 int tmp;
 
 tmp = firstEndofct(a,n);
 /*@ loop invariant tmp != 0 ==> is_endofct(a,n);
   @ loop assigns a[0..n-1],tmp; */
 while (tmp != 0 && isRGF(a,n) == 0) {
  tmp = nextEndofct(a,n);
 }
 if (tmp == 0) { return 0; }
 return 1;
}
