/* Generation of all permutations on [0..n-1], tabulated in an array [0..n-1]
 * in the increasing lexicographic order induced by the precedence order < on
 * integers. Here, by filtering among all endofunctions on [0..n-1] which are
 * not injections. */

#include "../fct/fct.h"
#include "../endofct/endofct.h"

/*@ predicate is_diff(int *a, integer n, integer i) =
  @  \forall integer j; 0 <= j < n ==> (a[j] == a[i] ==> j == i);
  @ predicate is_linear_gen(int *a, integer n, integer b) =
  @  \forall integer i; 0 <= i < b ==> is_diff(a,n,i);
  @ predicate is_linear(int *a, integer n) =
  @  \forall integer i; 0 <= i < n ==> is_diff(a,n,i);
  @ predicate is_injEndo(int *a, integer n) = 
     is_linear(a,n) && is_endofct(a,n); */

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires 0 <= i < n;
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_diff(a,n,i); */
int isDiff(int a[], int n, int i) {
 int j;

 /*@ loop invariant 0 <= j <= n;
   @ loop invariant is_diff(a,j,i);
   @ loop assigns j; */
 for (j = 0; j < n; j++) {
  if (a[j] == a[i] && j != i) {
   return 0;
  }
 }
 return 1;
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_linear(a,n); */
int isLinear(int a[], int n) {
 int i;

 /*@ loop invariant 0 <= i <= n;
   @ loop invariant is_linear_gen(a,n,i);
   @ loop assigns i; */
 for (i = 0; i < n; i++) {
  if (isDiff(a,n,i) != 1) {
   return 0;
  }
 }
 return 1;
}
   
/** Returns 1 when the array a is an injection, and 0 otherwise. */
/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires is_endofct(a,n);
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_injEndo(a,n); */
int isInjEndo(int a[], int n) {
 return (isLinear(a,n));
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_injEndo(a,n); */
int firstInjEndo(int a[], int n) {
 int tmp;
  
 tmp = firstEndofct(a,n);
 if (tmp != 1) { return 0; }
 
 /*@ loop invariant tmp != 0 ==> is_endofct(a,n);
   @ loop assigns a[0..n-1],tmp; */
 while (tmp != 0 && isInjEndo(a,n) == 0) {
  tmp = nextEndofct(a,n);
 } 
 
 if (tmp == 0) { return 0; }
 return 1;
}
 
/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires is_injEndo(a,n);
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_injEndo(a,n); */
int nextInjEndo(int a[], int n) {
 int tmp;
 
 /*@ loop invariant is_endofct(a,n);
   @ loop assigns a[0..n-1],tmp; */
 do {
  tmp = nextEndofct(a,n);
 } while (tmp != 0 && isInjEndo(a,n) == 0);

 if (tmp == 0) { return 0; }
 return 1;
}
