/* Generation of all injections from [0..n-1] to [0..k-1], tabulated in an array
 * [0..n-1] containing values between 0 and k-1, in the increasing lexicographic 
 * order induced by the precedence order < on integers.
 * Here, by filtering among all functions from [0..n-1] to [0..k-1].
 * Proved with Frama-C Neon and Alt-Ergo 0.95.1 or CVC3. */

#include "../fct/fct.h"
#include "../perm/perm.h"

/*@ predicate is_inj(int *a, integer n, integer k) = 
     is_fct(a,n,k) && is_linear(a,n); */

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires 0 <= i < n;
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_diff(a,n,i); */
int isDiff(int a[], int n, int i) {
 int j;

 /*@ loop invariant 0 <= j <= n;
   @ loop invariant is_diff(a,j,i);
   @ loop assigns j; */
 for (j = 0; j < n; j++) {
  if (a[j] == a[i] && j != i) {
   return 0;
  }
 }
 return 1;
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_linear(a,n); */
int isLinear(int a[], int n) {
 int i;

 /*@ loop invariant 0 <= i <= n;
   @ loop invariant is_linear_gen(a,n,i);
   @ loop assigns i; */
 for (i = 0; i < n; i++) {
  if (isDiff(a,n,i) != 1) {
   return 0;
  }
 }
 return 1;
}

/* Returns 1 when the array a is an injection, and 0 otherwise. */
/*@ requires n > 0 && k >= n && \valid(a+(0..n-1));
  @ requires is_fct(a,n,k);
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_inj(a,n,k); */
int isInj(int a[], int n, int k) {
 return (isLinear(a,n));
}

/*@ requires n > 0 && k >= n && \valid(a+(0..n-1));
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_inj(a,n,k); */
int firstInj(int a[], int n, int k) {
 int tmp;
  
 tmp = firstFct(a,n,k);
 if (tmp != 1) { return 0; }
 
 /*@ loop invariant is_fct(a,n,k);
   @ loop assigns a[0..n-1],tmp; */
 while (tmp != 0 && isInj(a,n,k) == 0) {
  tmp = nextFct(a,n,k);
 } 
 
 if (tmp == 0) { return 0; }
 return 1;
}
 
/*@ requires n > 0 && k >= n && \valid(a+(0..n-1));
  @ requires is_inj(a,n,k);
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_inj(a,n,k); */
int nextInj(int a[], int n, int k) {
 int tmp;
 
 /*@ loop invariant is_fct(a,n,k);
   @ loop assigns a[0..n-1],tmp; */
 do {
  tmp = nextFct(a,n,k);
 } while (tmp != 0 && isInj(a,n,k) == 0);

 if (tmp == 0) { return 0; }
 return 1;
}
