/* File: allex.c.
 *
 * This file defines a pattern for the ACSL contract and C code of a Boolean
 * function corresponding to a hybrid predicate defining a property of an array
 * of given length n. It covers all the cases when the hybrid predicate is of
 * the form
 *  \forall integer i; 0 <= i < n ==> (\exists integer j; 0 <= j < n && phi),
 * for an array a of length n, where phi is a quantifier-free formula.
 *
 * The predicate is named is_X1(a,n). The corresponding Boolean function is
 * named isX1. The inner quantified formula
 *  (\exists integer j; 0 <= j < n && ..)
 * is named is_X2(a,n,i). This hybrid predicate is implemented by a Boolean
 * function isX2. The quantifier-free formula phi in the predicate is_X2(a,n,v1)
 * is named is_X3(a,n,v1,j). It is implemented by the Boolean function isX3.
 *
 * Assume that the Boolean function isX3 correctly implements the predicate
 * is_X3. Then the contracts of the other Boolean functions is automatically
 * proved with frama-C, WP and Alt-Ergo.
 *
 * For k = 1,2, the predicate is_Xk is generalized by the predicate is_Xk_gen
 * with one more parameter. This generalization is useful to specify the loop
 * invariant of the corresponding Boolean functions.
 *
 * Commands:
 *  frama-c -wp allex.c -wp-skip-fct isX3 */

/*@ axiomatic preds {
  @  predicate is_X3(int *a, integer n, integer v1, integer v2) reads a[0..n-1];
  @ }
  @ predicate is_X2_gen(int *a, integer n, integer v1, integer v2) = \exists integer i2; 0 <= i2 < v2 && is_X3(a,n,v1,i2);
  @ predicate is_X2(int *a, integer n, integer v1) = is_X2_gen(a,n,v1,n);
  @ predicate is_X1_gen(int *a, integer n, integer v1) = \forall integer i1; 0 <= i1 < v1 ==> is_X2(a,n,i1);
  @ predicate is_X1(int *a, integer n) = is_X1_gen(a,n,n); */



/*@ requires n >= 0 && \valid(a+(0..n-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_X3(a,n,v1,v2); */
int isX3(int a[], int n, int v1, int v2);

/*@ requires n >= 0 && \valid(a+(0..n-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_X2(a,n,v1); */
int isX2(int a[], int n, int v1) {
 int i;

 /*@ loop invariant 0 <= i <= n;
   @ loop invariant ! is_X2_gen(a,n,v1,i);
   @ loop assigns i;
   @ loop variant n-i; */
 for (i = 0; i < n; i++)
  if (isX3(a,n,v1,i) == 1) return 1;
 return 0;
}

/*@ requires n >= 0 && \valid(a+(0..n-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_X1(a,n); */
int isX1(int a[], int n) {
 int i;

 /*@ loop invariant 0 <= i <= n;
   @ loop invariant is_X1_gen(a,n,i);
   @ loop assigns i;
   @ loop variant n-i; */
 for (i = 0; i < n; i++)
  if (isX2(a,n,i) == 0) return 0;
 return 1;
}
