/* File: exall.c.
 *
 * This file defines a pattern for the ACSL contract and C code of a Boolean
 * function corresponding to a hybrid predicate defining a property of an array
 * of given length. It covers all the cases when the hybrid predicate is of the
 * form
 *  \exists integer i; 0 <= i < n ==> (\forall integer j; 0 <= j < n && ..),
 * for an array a of length n.
 *
 * The predicate is named is_X1(a,n). The corresponding Boolean function is
 * named isX1. The inner quantified formula
 *  (\forall integer j; 0 <= j < n && ..)
 * is named is_X2(a,n,i). This hybrid predicate is implemented by a Boolean
 * function isX2. The quantifier-free formula in the predicate is_X2(a,n1,n2)
 * is named is_X3(a,n1,n2,j). It is implemented by the Boolean function isX3.
 *
 * Assume that the Boolean function isX3 correctly implements the predicate
 * is_X3. Then the contracts of the other Boolean functions is automatically
 * proved with frama-C, WP and Alt-Ergo.
 *
 * For k = 1,2, the predicate is_Xk is generalized by the predicate is_Xk_gen
 * with one more parameter. This generalization is useful to specify the loop
 * invariant of the corresponding Boolean functions.
 *
 * Commands:
 *  frama-c -wp exall.c -wp-fct isX1
 *  frama-c -wp exall.c -wp-fct isX2 */

/*@ axiomatic preds {
  @  predicate is_X3(int *a, integer n1, integer n2, integer n3) reads a[0..n1-1];
  @ }
  @
  @ predicate is_X2_gen(int *a, integer n1, integer n2, integer n3) =
  @  \forall integer i2; 0 <= i2 < n3 ==> is_X3(a,n1,n2,i2);
  @ predicate is_X2(int *a, integer n1, integer n2) = is_X2_gen(a,n1,n2,n1);
  @
  @ predicate is_X1_gen(int *a, integer n1, integer n2) =
  @  \exists integer i1; 0 <= i1 < n2 && is_X2(a,n1,i1);
  @ predicate is_X1(int *a, integer n) = is_X1_gen(a,n,n); */

/*@ requires 0 <= n1 && \valid(a+(0..n1-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_X3(a,n1,n2,n3); */
int isX3(int a[], int n1, int n2, int n3);

/*@ requires 0 <= n1 && \valid(a+(0..n1-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_X2(a,n1,n2); */
int isX2(int a[], int n1, int n2) {
 int i;

 /*@ loop invariant 0 <= i <= n1;
   @ loop invariant is_X2_gen(a,n1,n2,i);
   @ loop assigns i;
   @ loop variant n1-i; */
 for (i = 0; i < n1; i++) {
  if (isX3(a,n1,n2,i) == 0) {
   return 0;
  }
 }
 return 1;
}

/*@ requires n1 >= 0 && \valid(a+(0..n1-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_X1(a,n1);  */
int isX1(int a[], int n1) {
 int i;

 /*@ loop invariant 0 <= i <= n1;
   @ loop invariant ! is_X1_gen(a,n1,i);
   @ loop assigns i;
   @ loop variant n1-i; */
 for (i = 0; i < n1; i++) {
  if (isX2(a,n1,i) == 1) {
   return 1;
  }
 }
 return 0;
}
