/* Generation of all permutations of a set of n elements in lexicographic order.
 * Here, by filtering among all functions from [0..n-1] to [0..n-1]. */

#include "../fct/fct.h"

/*@ predicate is_diff(int *a, integer n, integer i) =
  @  \forall integer j; 0 <= j < n ==> (a[j] == a[i] ==> j == i);
  @ predicate is_linear_gen(int *a, integer n, integer b) =
  @  \forall integer i; 0 <= i < b ==> is_diff(a,n,i);
  @ predicate is_linear(int *a, integer n) =
  @  \forall integer i; 0 <= i < n ==> is_diff(a,n,i);
  @ predicate is_perm(int *a, integer n) = is_linear(a,n) && is_fct(a,n,n); */
  
/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires 0 <= i < n;
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_diff(a,n,i); */
int isDiff(int a[], int n, int i) {
 int j;

 /*@ loop invariant 0 <= j <= n;
   @ loop invariant is_diff(a,j,i);
   @ loop assigns j; */
 for (j = 0; j < n; j++) {
  if (a[j] == a[i] && j != i) {
   return 0;
  }
 }
 return 1;
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 <==> is_linear(a,n); */
int isLinear(int a[], int n) {
 int i;

 /*@ loop invariant 0 <= i <= n;
   @ loop invariant is_linear_gen(a,n,i);
   @ loop assigns i; */
 for (i = 0; i < n; i++) {
  if (isDiff(a,n,i) != 1) {
   return 0;
  }
 }
 return 1;
}

/** Returns 1 when the array a is a permutation, and 0 otherwise. */
/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires is_fct(a,n,n);
  @ assigns \nothing;
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_perm(a,n); */
int isPerm(int a[], int n) {
 return (isLinear(a,n));
}

/*@ requires n > 0 && \valid(a+(0..n-1));
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result != 0 ==> is_perm(a,n); */
int firstPerm(int a[], int n) {
 int tmp;
 
 tmp = firstFct(a,n,n);
 if (tmp == 0) { return 0; }
 
 /*@ loop invariant is_fct(a,n,n);
   @ loop assigns a[0..n-1],tmp; */
 do {
  tmp = nextFct(a,n,n);
 } while (tmp != 0 && isPerm(a,n) == 0);
 
 if (tmp == 0) { return 0; }
 return 1;
}
 
/*@ requires n > 0 && \valid(a+(0..n-1));
  @ requires is_perm(a,n);
  @ assigns a[0..n-1];
  @ ensures \result == 0 || \result == 1;
  @ ensures \result == 1 ==> is_perm(a,n); */
int nextPerm(int a[], int n) {
 int tmp;
 
 /*@ loop invariant is_fct(a,n,n);
   @ loop assigns a[0..n-1],tmp; */
 do {
  tmp = nextFct(a,n,n);
 } while (tmp != 0 && isPerm(a,n) == 0);

 if (tmp == 0) { return 0; }
 return 1;
}
